/* Wejp's ArmaSpy v0.0.3                                   */
/* Copyright (c) 2006 Johannes Heimansberg (wejp.mooo.com) */
/* Released under the GNU General Public License Version 2 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
#include "armaserver.h"
#include <gtk/gtk.h>
#include "eggtrayicon.h"
#include "arma_mini.xpm"
#define VERSION_NUMBER "0.0.3"
#define MAX_SERVERS 128
#define SERVER_URL "http://armagetron.de/data/frommaster.nfo"
/*#define SERVER_URL "http://www.armagetron.be/frommaster.nfo.crash"*/

static ArmaServer *server[MAX_SERVERS];

struct MemoryStruct {
	char   *memory;
	size_t  size;
};

void *myrealloc(void *ptr, size_t size)
{
	if (ptr)
		return realloc(ptr, size);
	else
		return malloc(size);
}

size_t cb_write_memory(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct MemoryStruct *mem = (struct MemoryStruct *)data;

	mem->memory = (char *)myrealloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
}

static void cb_quit_function(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

static void show_about_dialog(void)
{
	GtkWidget *dialog;

	dialog = gtk_about_dialog_new();
	gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(dialog), 
	                             VERSION_NUMBER);
	gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG(dialog),
	                               "Copyright (c) 2006 Johannes Heimansberg");
	gtk_about_dialog_set_comments(GTK_ABOUT_DIALOG(dialog),
	                               "Released under " \
	                               "the GNU General Public License V2.");
	gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(dialog),
	                             "http://wejp.mooo.com/");
	gtk_dialog_run(GTK_DIALOG(dialog));
}

static void cb_docklet_menu()
{
	GtkWidget *menu, *submenu_players[MAX_SERVERS], *entry;
	int        i, j, k = 0;
	gchar      entry_str[80];
	ArmaServer *server_sort[MAX_SERVERS], *server_unsort[MAX_SERVERS];

	menu = gtk_menu_new();

	for (i = 0; i < MAX_SERVERS; i++) {
		server_sort[i] = NULL;
		server_unsort[i] = server[i];
	}

	/* sort by players online */
	for (i = 0; i < MAX_SERVERS; i++) {
		for (j = 0; j < MAX_SERVERS; j++) {
			if (server_unsort[j] != NULL) {
				if (server_sort[i] == NULL) server_sort[i] = server_unsort[j];
				if (armaserver_get_players_online(server_unsort[j]) >= 
				    armaserver_get_players_online(server_sort[i])) {

					server_sort[i] = server_unsort[j];
					k = j;
				}
			}
		}
		server_unsort[k] = NULL;
	}

	for (i = 0; server_sort[i] != NULL && i < MAX_SERVERS; i++) {
		int dup = FALSE;
		if (armaserver_get_players_online(server_sort[i]) > 0) {
			submenu_players[i] = gtk_menu_new();
			for (j = 0; j < armaserver_get_players_online(server_sort[i]); j++) {
				entry = gtk_menu_item_new_with_label(armaserver_get_player_name(server_sort[i], j));
				gtk_menu_shell_append(GTK_MENU_SHELL(submenu_players[i]), entry);
			}

			for (j = 0; j < i; j++) {
				if (strcmp(armaserver_get_server_name(server_sort[i]),
			               armaserver_get_server_name(server_sort[j])) == 0) {
					dup = TRUE;
				}
			}
			if (dup == FALSE) {
				snprintf(entry_str, 80, "%s (%d players)",
				         armaserver_get_server_name(server_sort[i]),
				         armaserver_get_players_online(server_sort[i]));
				entry = gtk_menu_item_new_with_label(entry_str);
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(entry), submenu_players[i]);
				gtk_menu_shell_append(GTK_MENU_SHELL(menu), entry);
			}
		}
	}

	entry = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), entry);

	entry = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), entry);
	g_signal_connect(G_OBJECT(entry), "activate",
	                 G_CALLBACK(show_about_dialog), NULL);

	entry = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), entry);
	g_signal_connect(G_OBJECT(entry), "activate",
	                 G_CALLBACK(cb_quit_function), NULL);

	gtk_widget_show_all(menu);
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 0, 
	               gtk_get_current_event_time());
}

static void cb_docklet_clicked(GtkWidget      *widget,
                               GdkEventButton *event,
                               gpointer        data)
{
	if (event->type == GDK_BUTTON_PRESS) {
		switch (event->button) {
		case 1:
			break;
		case 3:
			cb_docklet_menu();
			break;
		}
	}
}

static gboolean cb_timer(gpointer userdata)
{
	CURL               *curl_handle;
	int                 i, j;
	char                server_name[64], player[MAX_PLAYERS][64], buf[64];
	int                 servers = 0, players = 0, state = 0, player_cnt = 0;
	struct MemoryStruct chunk;

	enum State { SERVER_NAME = 1, NUMBER_OF_PLAYERS = 2, PLAYER_NAMES = 3 };

	printf("Fetching data...\n");
	for (i = 0; server[i] != NULL && i < MAX_SERVERS; i++)
		g_free(server[i]);
	for (i = 0; i < MAX_SERVERS; i++)
		server[i] = NULL;
	for (i = 0; i < MAX_PLAYERS; i++) {
		player[i][63] = '\0';
		for (j = 0; j < 63; j++)
			player[i][j] = ' ';
	}

	for (i = 0; i < 63; i++)
		buf[i] = ' ';
	buf[63] = '\0';

	chunk.memory = NULL; /* we expect realloc(NULL, size) to work */
	chunk.size   = 0;    /* no data at this point */

	curl_global_init(CURL_GLOBAL_ALL);

	curl_handle = curl_easy_init();
	curl_easy_setopt(curl_handle, CURLOPT_URL, SERVER_URL);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, cb_write_memory);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)&chunk);
	sprintf(buf, "armaspy/%s", VERSION_NUMBER);
	curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, buf);
	curl_easy_perform(curl_handle);
	curl_easy_cleanup(curl_handle);

	state = SERVER_NAME;
	for (i = 0, j = 0; i < chunk.size && servers <= MAX_SERVERS; i++) {
		if (chunk.memory[i] >= 8) { /* common character */
			/* read next character */
			if (j < 63) buf[j] = chunk.memory[i];
			j++;
		} else if (chunk.memory[i] < 8) { /* separator */
			switch (state) {
				case SERVER_NAME:
					buf[j] = '\0';
					strncpy(server_name, buf, 63);
					player_cnt = 0;
					state = NUMBER_OF_PLAYERS;
					break;
				case NUMBER_OF_PLAYERS:
					buf[j] = '\0';
					players = atoi(buf);
					if (players > MAX_PLAYERS) players = MAX_PLAYERS;
					state = PLAYER_NAMES;
					break;
				case PLAYER_NAMES:
					if (chunk.memory[i] == '\2') { /* another player name follows */
						buf[j] = '\0';
						strncpy(player[player_cnt], buf, 63);
						if (player_cnt < MAX_PLAYERS) player_cnt++;
					} else if (chunk.memory[i] == '\3') { /* last player name reached */
						int k;
						/*printf("players found: %d, players expected: %d\n", player_cnt, players);*/
						server[servers] = g_malloc(sizeof(ArmaServer));
						armaserver_init(server[servers], server_name);
						for (k = 0; k < players; k++)
							armaserver_add_player(server[servers], player[k]);
						servers++;
						state = SERVER_NAME;
					}
					break;
			}
			j = 0;
		}
	}

	if (chunk.memory)
		free(chunk.memory);

	return TRUE;
}

int main(int argc, char **argv)
{
	GtkWidget   *vbox, *image_docklet;
	EggTrayIcon *docklet;
	GtkTooltips *docklet_tooltip;
	GdkPixbuf   *icon;
	int          i;

	gtk_init(&argc, &argv);

	/* tray icon */
	docklet = egg_tray_icon_new("Wejp's ArmaSpy");
	vbox = gtk_event_box_new();
	icon = gdk_pixbuf_new_from_xpm_data((const char **)arma_mini_xpm);
	image_docklet = gtk_image_new_from_pixbuf(icon);

	g_signal_connect(G_OBJECT(vbox), "button-press-event",
	                 G_CALLBACK(cb_docklet_clicked), NULL);

	gtk_container_add(GTK_CONTAINER(vbox), image_docklet);
	gtk_container_add(GTK_CONTAINER(docklet), vbox);
	gtk_widget_show_all(GTK_WIDGET(docklet));
	
	docklet_tooltip = gtk_tooltips_new();
	gtk_tooltips_set_tip(docklet_tooltip, GTK_WIDGET(docklet),
	                     "Wejp's ArmaSpy", NULL);
	gtk_tooltips_enable(docklet_tooltip);

	printf("Wejp's ArmaSpy %s (http://wejp.mooo.com)\n", VERSION_NUMBER);
	g_timeout_add(90000, &cb_timer, NULL);
	cb_timer(NULL);
	gtk_main();

	for (i = 0; server[i] != NULL; i++)
		free(server[i]);
	return 0;
}
